@@ -1,5 +1,8 @@
 import copy
 
+from rest_framework.fields import (
+    empty, ListField, DictField
+)
 from rest_framework.serializers import (
     Serializer, ListSerializer, SerializerMethodField,
     ValidationError, PrimaryKeyRelatedField
@@ -13,33 +16,41 @@ from .operations import ADD, CREATE, REMOVE, UPDATE
 CREATE_SUPPORTED_OPERATIONS = (ADD, CREATE)
 UPDATE_SUPPORTED_OPERATIONS = (ADD, CREATE, REMOVE, UPDATE)
 
+
 class DynamicSerializerMethodField(SerializerMethodField):
     def to_representation(self, value):
         method = getattr(self.parent, self.method_name)
-        if self.field_name in self.parent.nested_fields:
+        if (hasattr(self.parent, "nested_fields") and 
+                self.field_name in self.parent.nested_fields):
             query = self.parent.nested_fields[self.field_name]
         else:
+            # Include all fields
             query = {
-                "include": [],
-                "exclude": []
+                "include": ["*"],
+                "exclude": [],
+                "arguments": {}
             }
         return method(value, query)
 
 
-class _ReplaceableField(object):
-    pass
+class BaseReplaceableNestedField(object):
+    def to_internal_value(self, data):
+        raise NotImplementedError('`to_internal_value()` must be implemented.')
 
-class _WritableField(object):
-    pass
 
-def BaseNestedFieldSerializerFactory(*args, 
-                                     accept_pk=False, 
-                                     create_ops=[ADD, CREATE], 
-                                     update_ops=[ADD, CREATE, REMOVE, UPDATE], 
-                                     serializer_class=None, 
-                                     **kwargs):
-    BaseClass = _ReplaceableField if accept_pk else _WritableField
-    
+class BaseWritableNestedField(object):
+    def to_internal_value(self, data):
+        raise NotImplementedError('`to_internal_value()` must be implemented.')
+
+
+def BaseNestedFieldSerializerFactory(
+        *args,
+        partial=None,
+        accept_pk=False,
+        create_ops=[ADD, CREATE],
+        update_ops=[ADD, CREATE, REMOVE, UPDATE],
+        serializer_class=None,
+        **kwargs):
     if not set(create_ops).issubset(set(CREATE_SUPPORTED_OPERATIONS)):
         msg = (
             "Invalid create operation, Supported operations are " +
@@ -53,23 +64,31 @@ def BaseNestedFieldSerializerFactory(*args,
             ", ".join(UPDATE_SUPPORTED_OPERATIONS)
         )
         raise InvalidOperation(msg)
+    
+    BaseClass = BaseReplaceableNestedField if accept_pk \
+        else BaseWritableNestedField
 
-    class BaseNestedFieldListSerializer(ListSerializer, BaseClass):
+    class BaseNestedField(BaseClass):
         @property
         def is_partial(self):
-            request = self.context.get('request')
-            partial = True if request.method == "PATCH" else False
-            return partial
+            if partial is None and self.parent is not None:
+                return self.parent.partial
+            else:
+                return partial
+
+    class BaseNestedFieldListSerializer(ListSerializer, BaseNestedField):
 
         def validate_pk_list(self, pks):
+            ListField().run_validation(pks)
             queryset = self.child.Meta.model.objects.all()
             validator = PrimaryKeyRelatedField(
-                queryset=queryset, 
-                many=True 
+                queryset=queryset,
+                many=True
             )
             return validator.run_validation(pks)
     
-        def validate_data_list(self, data, partial=False):
+        def validate_data_list(self, data):
+            ListField().run_validation(data)
             model = self.parent.Meta.model
             rel = getattr(model, self.field_name).rel
 
@@ -82,9 +101,10 @@ def BaseNestedFieldSerializerFactory(*args,
                 original_fields = copy.copy(serializer_class.Meta.fields)
                 serializer_class.Meta.fields = list(fields)
                 parent_serializer = serializer_class(
-                    data=data, 
-                    many=True, 
-                    partial=partial,
+                    **self.child.validation_kwargs,
+                    data=data,
+                    many=True,
+                    partial=self.is_partial,
                     context=self.context
                 )
                 parent_serializer.is_valid(raise_exception=True)
@@ -92,9 +112,10 @@ def BaseNestedFieldSerializerFactory(*args,
             else:
                 # ManyToMany Relation
                 parent_serializer = serializer_class(
-                    data=data, 
-                    many=True, 
-                    partial=partial,
+                    **self.child.validation_kwargs,
+                    data=data,
+                    many=True,
+                    partial=self.is_partial,
                     context=self.context
                 )
                 parent_serializer.is_valid(raise_exception=True)
@@ -104,75 +125,79 @@ def BaseNestedFieldSerializerFactory(*args,
             return self.validate_pk_list(data)
 
         def validate_create_list(self, data):
-            return self.validate_data_list(data, partial=False)
+            return self.validate_data_list(data)
     
         def validate_remove_list(self, data):
             return self.validate_pk_list(data)
     
         def validate_update_list(self, data):
-            # Obtain pks & data then
-            if isinstance(data, dict):
-                self.validate_pk_list(data.keys())
-                values = list(data.values())
-                self.validate_data_list(values, partial=self.is_partial)
-            else:
-                raise ValidationError(
-                    "Expected data of form {'pk': 'data'..}"
-                )
-
-        def create_data_is_valid(self, data):
-            if (isinstance(data, dict) and 
-                    set(data.keys()).issubset(create_ops)):
-                return True
-            return False
+            DictField().run_validation(data)
+            pks = list(data.keys())
+            self.validate_pk_list(pks)
+            values = list(data.values())
+            self.validate_data_list(values)
 
         def data_for_create(self, data):
             validate = {
                 ADD: self.validate_add_list,
-                CREATE: self.validate_create_list, 
+                CREATE: self.validate_create_list,
             }
 
-            if self.create_data_is_valid(data):
-                for operation, values in data.items():
+            DictField().run_validation(data)
+            for operation, values in data.items():
+                try:
                     validate[operation](values)
-                return data
-            else:
-                wrap_quotes = lambda op: "'" + op + "'"
-                op_list =list(map(wrap_quotes, create_ops))
-                msg = (
-                    "Expected data of form " +
-                    "{" + ": [..], ".join(op_list) + ": [..]}"
-                )
-                raise ValidationError(msg)
-
-        def update_data_is_valid(self, data):
-            if (isinstance(data, dict) and 
-                    set(data.keys()).issubset(update_ops)):
-                return True
-            return False
+                except ValidationError as e:
+                    detail = {operation: e.detail}
+                    code = e.get_codes()
+                    raise ValidationError(detail, code) from None
+                except KeyError:
+                    wrap_quotes = lambda op: "`" + op + "`"
+                    op_list =list(map(wrap_quotes, create_ops))
+                    msg = (
+                        "`%s` is not a valid operation, valid operations "
+                        "for this request are %s" 
+                        % (operation, ', '.join(op_list))
+                    )
+                    code = 'invalid_operation'
+                    raise ValidationError(msg, code=code) from None
+            return data
 
         def data_for_update(self, data):
             validate = {
                 ADD: self.validate_add_list,
-                CREATE: self.validate_create_list, 
-                REMOVE: self.validate_remove_list, 
+                CREATE: self.validate_create_list,
+                REMOVE: self.validate_remove_list,
                 UPDATE: self.validate_update_list,
             }
 
-            if self.update_data_is_valid(data):
-                for operation, values in data.items():
+            DictField().run_validation(data)
+            for operation, values in data.items():
+                try:
                     validate[operation](values)
-                return data
-            else:
-                wrap_quotes = lambda op: "'" + op + "'"
-                op_list = list(map(wrap_quotes, update_ops))
-                msg = (
-                    "Expected data of form " +
-                    "{" + ": [..], ".join(op_list) + ": [..]}"
-                )
-                raise ValidationError(msg)
+                except ValidationError as e:
+                    detail = {operation: e.detail}
+                    code = e.get_codes()
+                    raise ValidationError(detail, code) from None
+                except KeyError:
+                    wrap_quotes = lambda op: "`" + op + "`"
+                    op_list =list(map(wrap_quotes, update_ops))
+                    msg = (
+                        "`%s` is not a valid operation, valid operations "
+                        "for this request are %s" 
+                        % (operation, ', '.join(op_list))
+                    )
+                    code='invalid_operation'
+                    raise ValidationError(msg, code=code) from None
+            return data
 
         def to_internal_value(self, data):
+            source = kwargs.get('source', None)
+            if source is not None and self.parent is not None:
+                alias = self.field_name
+                self.parent._aliased_nested_fields.update({source: self})
+                self.parent._unaliased_fields.pop(alias)
+            required = kwargs.get('required', True)
             request = self.context.get('request')
             if request.method in ["PUT", "PATCH"]:
                 return self.data_for_update(data)
@@ -181,8 +206,9 @@ def BaseNestedFieldSerializerFactory(*args,
                 return self.data_for_create(data)
 
             parent_serializer = serializer_class(
-                data=data, 
-                many=True, 
+                **self.child.validation_kwargs,
+                data=data,
+                many=True,
                 partial=self.is_partial,
                 context=self.context
             )
@@ -195,16 +221,10 @@ def BaseNestedFieldSerializerFactory(*args,
                 (serializer_class.__name__, )
             )
 
-    class BaseNestedFieldSerializer(serializer_class, BaseClass):
+    class BaseNestedFieldSerializer(serializer_class, BaseNestedField):
         class Meta(serializer_class.Meta):
             list_serializer_class = BaseNestedFieldListSerializer
 
-        @property
-        def is_partial(self):
-            request = self.context.get('request')
-            partial = True if request.method == "PATCH" else False
-            return partial
-
         def run_validation(self, data):
             # Run `to_internal_value` only nothing more
             # This is needed only on DRF 3.8.x due to a bug on it
@@ -215,6 +235,7 @@ def BaseNestedFieldSerializerFactory(*args,
         def validate_pk_based_nested(self, data):
             queryset = self.Meta.model.objects.all()
             validator = PrimaryKeyRelatedField(
+                **self.validation_kwargs,
                 queryset=queryset,
                 many=False
             )
@@ -223,7 +244,8 @@ def BaseNestedFieldSerializerFactory(*args,
 
         def validate_data_based_nested(self, data):
             parent_serializer = serializer_class(
-                data=data, 
+                **self.validation_kwargs,
+                data=data,
                 partial=self.is_partial,
                 context=self.context
             )
@@ -231,6 +253,26 @@ def BaseNestedFieldSerializerFactory(*args,
             return parent_serializer.validated_data
 
         def to_internal_value(self, data):
+            source = kwargs.get('source', None)
+            if source is not None and self.parent is not None:
+                alias = self.field_name
+                self.parent._aliased_nested_fields.update({source: self})
+                self.parent._unaliased_fields.pop(alias)
+            required = kwargs.get('required', True)
+            default = kwargs.get('default', empty)
+
+            if data == empty and required and default == empty:
+                raise ValidationError(
+                    "This field is required.",
+                    code='required'
+                )
+            elif data == empty and required:
+                data = default
+            elif data == empty and default == empty:
+                data = ""
+                
+            if data == "":
+                data = None
             if accept_pk:
                 return self.validate_pk_based_nested(data)
             return self.validate_data_based_nested(data)
@@ -241,7 +283,9 @@ def BaseNestedFieldSerializerFactory(*args,
                 (serializer_class.__name__, )
             )
 
-    kwargs.update({"read_only": False, "write_only": False})
+    read_only = kwargs.get('read_only', False)
+    write_only = kwargs.get('write_only', False)
+    kwargs.update({"read_only": read_only, "write_only": write_only})
     return {
         "serializer_class": BaseNestedFieldSerializer,
         "list_serializer_class": BaseNestedFieldListSerializer,
@@ -254,14 +298,26 @@ def NestedFieldWraper(*args, **kwargs):
     factory = BaseNestedFieldSerializerFactory(*args, **kwargs)
     serializer_class = kwargs["serializer_class"]
 
+    serializer_validation_kwargs = {**factory['kwargs']}
+    # TODO: Find all non validation kwargs to remove(below are just few)
+
+    # Remove non validation related kwargs from `valdation_kwargs`
+    non_validation_kwargs = ['many', 'data', 'instance', 'context']
+    for kwarg in non_validation_kwargs:
+        if kwarg in serializer_validation_kwargs:
+            serializer_validation_kwargs.pop(kwarg)
+
     class NestedListSerializer(factory["list_serializer_class"]):
         def __repr__(self):
             return (
                 "NestedField(%s, many=False)" % 
                 (serializer_class.__name__, )
             )
-
+            
     class NestedSerializer(factory["serializer_class"]):
+        # set validation related kwargs to be used on 
+        # `NestedCreateMixin` and `NestedUpdateMixin`
+        validation_kwargs = serializer_validation_kwargs
         class Meta(factory["serializer_class"].Meta):
             list_serializer_class = NestedListSerializer
 
@@ -279,8 +335,7 @@ def NestedFieldWraper(*args, **kwargs):
 
 def NestedField(serializer_class, *args, **kwargs):
     return NestedFieldWraper(
-        *args, 
-        serializer_class=serializer_class, 
-        **kwargs
+        *args,
+        **kwargs,
+        serializer_class=serializer_class
     )
-
