@@ -1,31 +1,69 @@
 import copy
+import collections
 
 from rest_framework.serializers import (
     Serializer, ListSerializer,
     ValidationError
 )
+from .settings import restql_settings
+from django.db.models import Prefetch
 from django.db.models.fields.related import(
     ManyToOneRel, ManyToManyRel
 )
 
 from .parser import Parser
-from .exceptions import FieldNotFound
+from .exceptions import FieldNotFound, QueryFormatError
 from .operations import ADD, CREATE, REMOVE, UPDATE
 from .fields import (
-    _ReplaceableField, _WritableField,
+    BaseReplaceableNestedField, BaseWritableNestedField,
     DynamicSerializerMethodField
 )
 
 
-class DynamicFieldsMixin(object):
-    query_param_name = "query"
+class RequestQueryParserMixin(object):
+    @staticmethod
+    def get_restql_query_param_name():
+        DEFAULT_QUERY_PARAM_NAME = 'query'
+        query_param_name = getattr(
+            restql_settings,
+            "QUERY_PARAM_NAME",
+            DEFAULT_QUERY_PARAM_NAME
+        )
+        return query_param_name
 
+    @classmethod
+    def has_restql_query_param(cls, request):
+        query_param_name = cls.get_restql_query_param_name()
+        return query_param_name in request.query_params
+
+    @classmethod
+    def get_raw_restql_query(cls, request):
+        query_param_name = cls.get_restql_query_param_name()
+        return request.query_params[query_param_name]
+
+    @classmethod
+    def get_parsed_restql_query_from_req(cls, request):
+        if hasattr(request, 'parsed_restql_query'):
+            return request.parsed_restql_query
+        raw_query = cls.get_raw_restql_query(request)
+        parser = Parser(raw_query)
+        parsed_restql_query = parser.get_parsed()
+
+        # Save parsed restql query to the request so that
+        # we won't need to parse it again if needed
+        request.parsed_restql_query = parsed_restql_query
+        return parsed_restql_query
+
+
+class DynamicFieldsMixin(RequestQueryParserMixin):
     def __init__(self, *args, **kwargs):
-        # Don't pass 'query', 'fields' and 'exclude' kwargs to the superclass
-        self.query = kwargs.pop('query', None)  # Parsed query
+        # Don't pass 'query', 'fields', 'exclude', 'return_pk' 
+        # and 'disable_dynamic_fields'  kwargs to the superclass
+        self.parsed_restql_query = kwargs.pop('query', None)
         self.allowed_fields = kwargs.pop('fields', None)
         self.excluded_fields = kwargs.pop('exclude', None)
         self.return_pk = kwargs.pop('return_pk', False)
+        self.disable_dynamic_fields = kwargs.pop('disable_dynamic_fields', False)
 
         is_field_kwarg_set = self.allowed_fields is not None
         is_exclude_kwarg_set = self.excluded_fields is not None
@@ -40,29 +78,6 @@ class DynamicFieldsMixin(object):
             return instance.pk
         return super().to_representation(instance)
 
-    @classmethod
-    def has_query_param(cls, request):
-        return cls.query_param_name in request.query_params
-
-    @classmethod
-    def get_raw_query(cls, request):
-        return request.query_params[cls.query_param_name]
-
-    @classmethod
-    def get_parsed_query_from_req(cls, request):
-        raw_query = cls.get_raw_query(request)
-        parser = Parser(raw_query)
-        try:
-            parsed_query = parser.get_parsed()
-            return parsed_query
-        except SyntaxError as e:
-            msg = (
-                "QueryFormatError: " + 
-                e.msg + " on " + 
-                e.text
-            )
-            raise ValidationError(msg) from None
-
     def get_allowed_fields(self):
         fields = super().fields
         if self.allowed_fields is not None:
@@ -101,7 +116,7 @@ class DynamicFieldsMixin(object):
     @staticmethod
     def is_nested_field(field_name, field, raise_error=False):
         nested_classes = (
-            Serializer, ListSerializer, 
+            Serializer, ListSerializer,
             DynamicSerializerMethodField
         )
         if isinstance(field, nested_classes):
@@ -121,9 +136,10 @@ class DynamicFieldsMixin(object):
         # The format is  {nested_field: [sub_fields ...] ...}
         allowed_nested_fields = {}
 
-        # The self.query["include"] contains a list of allowed fields
+        # The self.parsed_restql_query["include"] 
+        # contains a list of allowed fields,
         # The format is [field, {nested_field: [sub_fields ...]} ...]
-        included_fields =  self.query["include"]
+        included_fields =  self.parsed_restql_query["include"]
         include_all_fields = False
         for field in included_fields:
             if field == "*":
@@ -134,12 +150,12 @@ class DynamicFieldsMixin(object):
                 # Nested field
                 for nested_field in field:
                     self.is_field_found(
-                        nested_field, 
+                        nested_field,
                         all_field_names,
                         raise_error=True
                     )
                     self.is_nested_field(
-                        nested_field, 
+                        nested_field,
                         all_fields[nested_field],
                         raise_error=True
                     )
@@ -171,9 +187,10 @@ class DynamicFieldsMixin(object):
         # The format is  {nested_field: [sub_fields ...] ...}
         allowed_nested_fields = {}
 
-        # The self.query["include"] contains a list of expanded nested fields
+        # The self.parsed_restql_query["include"] 
+        # contains a list of expanded nested fields
         # The format is [{nested_field: [sub_field]} ...]
-        nested_fields = self.query["include"]
+        nested_fields = self.parsed_restql_query["include"]
         for field in nested_fields:
             if field == "*":
                 # Ignore this since it's not an actual field(it's just a flag)
@@ -191,8 +208,9 @@ class DynamicFieldsMixin(object):
                 )
             allowed_nested_fields.update(field)
 
-        # self.query["exclude"] is a list of names of excluded fields
-        excluded_fields = self.query["exclude"]
+        # self.parsed_restql_query["exclude"] 
+        # is a list of names of excluded fields
+        excluded_fields = self.parsed_restql_query["exclude"]
         for field in excluded_fields:
             self.is_field_found(field, all_field_names, raise_error=True)
             all_fields.pop(field)
@@ -207,7 +225,8 @@ class DynamicFieldsMixin(object):
         is_not_a_request_to_process = (
             request is None or 
             request.method != "GET" or 
-            not self.has_query_param(request)
+            self.disable_dynamic_fields or
+            not self.has_restql_query_param(request)
         )
 
         if is_not_a_request_to_process:
@@ -224,68 +243,256 @@ class DynamicFieldsMixin(object):
         )
 
         if is_top_retrieve_request or is_top_list_request:
-            if self.query is None:
-                # Use a query from the request
-                self.query = self.get_parsed_query_from_req(request)
+            if self.parsed_restql_query is None:
+                # Use a parsed query from the request
+                try:
+                    self.parsed_restql_query = \
+                        self.get_parsed_restql_query_from_req(request)
+                except SyntaxError as e:
+                    msg = "QuerySyntaxError: " + e.msg + " on " + e.text
+                    raise ValidationError(msg) from None
+                except QueryFormatError as e:
+                    msg = "QueryFormatError: " + str(e)
+                    raise ValidationError(msg) from None
+
         elif isinstance(self.parent, ListSerializer):
             field_name = self.parent.field_name
             parent = self.parent.parent
             if hasattr(parent, "nested_fields"):
                 parent_nested_fields = parent.nested_fields
-                self.query = parent_nested_fields.get(field_name, None)
+                self.parsed_restql_query = \
+                    parent_nested_fields.get(field_name, None)
         elif isinstance(self.parent, Serializer):
             field_name = self.field_name
             parent = self.parent
             if hasattr(parent, "nested_fields"):
                 parent_nested_fields = parent.nested_fields
-                self.query = parent_nested_fields.get(field_name, None)
+                self.parsed_restql_query = \
+                    parent_nested_fields.get(field_name, None)
         else:
             # Unkown scenario
             # No filtering of fields
             return self.get_allowed_fields()
 
-        if self.query is None:
+        if self.parsed_restql_query is None:
             # No filtering on nested fields
             # Retrieve all nested fields
             return self.get_allowed_fields()
 
-        # NOTE: self.query["include"] not being empty is not a guarantee 
-        # that the exclude operator(-) has not been used because the same 
-        # self.query["include"] is used to store nested fields when the
-        # exclude operator(-) is used
-        if self.query["exclude"]:
+        # NOTE: self.parsed_restql_query["include"] not being empty 
+        # is not a guarantee that the exclude operator(-) has not been 
+        # used because the same self.parsed_restql_query["include"]
+        # is used to store nested fields when the exclude operator(-) is used
+        if self.parsed_restql_query["exclude"]:
             # Exclude fields from a query
             return self.exclude_fields()
-        elif self.query["include"]:
-            # Here we are sure that self.query["exclude"] is empty
-            # which means the exclude operator(-) is not used, so
-            # self.query["include"] contains only fields to include
+        elif self.parsed_restql_query["include"]:
+            # Here we are sure that self.parsed_restql_query["exclude"] 
+            # is empty which means the exclude operator(-) is not used,
+            # so self.parsed_restql_query["include"] contains only fields
+            # to include
             return self.include_fields()
-        
-        # No fields to include or exclude so return all fields
-        return self.get_allowed_fields()
+        else:
+            # The query is empty i.e query={}
+            # return nothing
+            return {}
 
 
-class NestedCreateMixin(object):
+class EagerLoadingMixin(RequestQueryParserMixin):
+    @property
+    def parsed_restql_query(self):
+        """
+        Gets parsed query for use in eager loading. 
+        Defaults to the serializer parsed query assuming
+        using django-restql DynamicsFieldMixin.
+        """
+        if self.has_restql_query_param(self.request):
+            try:
+                return self.get_parsed_restql_query_from_req(self.request)
+            except (SyntaxError, QueryFormatError):
+                pass
+
+        # Else include all fields
+        query = {
+            "include": ["*"],
+            "exclude": [],
+            "arguments": {}
+        }
+        return query
+
+    @property
+    def should_auto_apply_eager_loading(self):
+        if hasattr(self, 'auto_apply_eager_loading'):
+            return self.auto_apply_eager_loading
+        return getattr(
+            restql_settings,
+            "AUTO_APPLY_EAGER_LOADING",
+            True
+        )
+
+    def get_select_related_mapping(self):
+        if hasattr(self, "select_related"):
+            return self.select_related
+        # Else select nothing
+        return {}
+
+    def get_prefetch_related_mapping(self):
+        if hasattr(self, "prefetch_related"):
+            return self.prefetch_related
+        # Else prefetch nothing
+        return {}
+
+    @classmethod
+    def get_dict_parsed_restql_query(cls, parsed_restql_query):
+        """
+        Returns the parsed query as a dict.
+        """
+        keys = {}
+        include = parsed_restql_query.get("include", [])
+        exclude = parsed_restql_query.get("exclude", [])
+
+        for item in include:
+            if isinstance(item, str):
+                keys[item] = True
+            elif isinstance(item, dict):
+                for key, nested_items in item.items():
+                    key_base = key
+                    nested_keys = cls.get_dict_parsed_restql_query(nested_items)
+                    keys[key_base] = nested_keys
+
+        for item in exclude:
+            if isinstance(item, str):
+                keys[item] = False
+            elif isinstance(item, dict):
+                for key, nested_items in item.items():
+                    key_base = key
+                    nested_keys = cls.get_dict_parsed_restql_query(nested_items)
+                    keys[key_base] = nested_keys
+        return keys
+
+    @staticmethod
+    def get_related_fields(related_fields_mapping, dict_parsed_restql_query):
+        """
+        Returns only whitelisted related fields from a query to be used on
+        `select_related` and `prefetch_related`
+        """
+        related_fields = []
+        for key, related_field in related_fields_mapping.items():
+            fields = key.split(".")
+            if isinstance(related_field, (str, Prefetch)):
+                related_field = [related_field]
+
+            query_node = dict_parsed_restql_query
+            for field in fields:
+                if isinstance(query_node, dict):
+                    if field in query_node:
+                        # Get a more specific query node
+                        query_node = query_node[field]
+                    elif "*" in query_node:
+                        # All fields are included
+                        continue
+                    else:
+                        # The field is not included in a query so
+                        # don't include this field in `related_fields`
+                        break
+            else:
+                # If the loop completed without breaking
+                if isinstance(query_node, dict) or query_node:
+                    related_fields.extend(related_field)
+        return related_fields
+
+    def apply_eager_loading(self, queryset):
+        """
+        Applies appropriate select_related and prefetch_related calls on a
+        queryset
+        """
+        query = self.get_dict_parsed_restql_query(self.parsed_restql_query)
+        select_mapping = self.get_select_related_mapping()
+        prefetch_mapping = self.get_prefetch_related_mapping()
+
+        to_select = self.get_related_fields(select_mapping, query)
+        to_prefetch = self.get_related_fields(prefetch_mapping, query)
+
+        if to_select:
+            queryset = queryset.select_related(*to_select)
+        if to_prefetch:
+            queryset = queryset.prefetch_related(*to_prefetch)
+        return queryset
+
+    def get_eager_queryset(self, queryset):
+        return self.apply_eager_loading(queryset)
+
+    def get_queryset(self):
+        """
+        Override for DRF's get_queryset on the view. 
+        If get_queryset is not present, we don't try to run this. 
+        Instead, this can still be used by manually calling 
+        self.get_eager_queryset and passing in the queryset desired.
+        """
+        if hasattr(super(), "get_queryset"):
+            queryset = super().get_queryset()
+            if self.should_auto_apply_eager_loading:
+                queryset = self.get_eager_queryset(queryset)
+            return queryset
+
+
+class BaseNestedMixin(object):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._unaliased_fields = self.get_fields()
+        self._aliased_nested_fields = {}
+
+    def _get_all_fields(self):
+        return collections.ChainMap(
+            self._unaliased_fields,
+            self._aliased_nested_fields
+        )
+
+    def to_internal_value(self, data):
+        return super().to_internal_value(data)
+
+
+class NestedCreateMixin(BaseNestedMixin):
     """ Create Mixin """
     def create_writable_foreignkey_related(self, data):
         # data format {field: {sub_field: value}}
         objs = {}
+        all_fields = self._get_all_fields()
         for field, value in data.items():
-            # Get serializer class for nested field
-            SerializerClass = type(self.get_fields()[field])
-            serializer = SerializerClass(data=value, context=self.context)
+            # Get nested field serializer
+            serializer = all_fields[field]
+            serializer_class = type(serializer)
+            kwargs = serializer.validation_kwargs
+            serializer = serializer_class(
+                **kwargs,
+                data=value,
+                context=self.context,
+                partial=serializer.is_partial
+            )
             serializer.is_valid()
-            obj = serializer.save()
-            objs.update({field: obj})
+            if value is None:
+                objs.update({field: None})
+            else:
+                obj = serializer.save()
+                objs.update({field: obj})
         return objs
 
     def bulk_create_objs(self, field, data):
-        model = self.get_fields()[field].child.Meta.model
-        SerializerClass = type(self.get_fields()[field].child)
+        all_fields = self._get_all_fields()
+        model = all_fields[field].child.Meta.model
+
+        # Get nested field serializer
+        serializer = all_fields[field].child
+        serializer_class = type(serializer)
+        kwargs = serializer.validation_kwargs
         pks = []
         for values in data:
-            serializer = SerializerClass(data=values, context=self.context)
+            serializer = serializer_class(
+                **kwargs,
+                data=values,
+                context=self.context,
+                partial=serializer.is_partial
+            )
             serializer.is_valid()
             obj = serializer.save()
             pks.append(obj.pk)
@@ -295,17 +502,18 @@ class NestedCreateMixin(object):
         # data format {field: {
         # foreignkey_name: name,
         # data: {
-        # ADD: [pks], 
+        # ADD: [pks],
         # CREATE: [{sub_field: value}]
         # }}
         field_pks = {}
         for field, values in data.items():
             model = self.Meta.model
             foreignkey = getattr(model, field).field.name
+            all_fields = self._get_all_fields()
             for operation in values:
                 if operation == ADD:
                     pks = values[operation]
-                    model = self.get_fields()[field].child.Meta.model
+                    model = all_fields[field].child.Meta.model
                     qs = model.objects.filter(pk__in=pks)
                     qs.update(**{foreignkey: instance.pk})
                     field_pks.update({field: pks})
@@ -318,19 +526,18 @@ class NestedCreateMixin(object):
 
     def create_many_to_many_related(self, instance, data):
         # data format {field: {
-        # ADD: [pks], 
+        # ADD: [pks],
         # CREATE: [{sub_field: value}]
         # }}
         field_pks = {}
         for field, values in data.items():
+            obj = getattr(instance, field)
             for operation in values:
                 if operation == ADD:
-                    obj = getattr(instance, field)
                     pks = values[operation]
                     obj.set(pks)
                     field_pks.update({field: pks})
                 elif operation == CREATE:
-                    obj = getattr(instance, field)
                     pks = self.bulk_create_objs(field, values[operation])
                     obj.set(pks)
                     field_pks.update({field: pks})
@@ -341,7 +548,7 @@ class NestedCreateMixin(object):
             "foreignkey_related": { 
                 "replaceable": {},
                 "writable": {}
-            }, 
+            },
             "many_to": {
                 "many_related": {},
                 "one_related": {}
@@ -351,20 +558,26 @@ class NestedCreateMixin(object):
         # Make a partal copy of validated_data so that we can
         # iterate and alter it
         data = copy.copy(validated_data)
+        all_fields = self._get_all_fields()
         for field in data:
-            field_serializer = self.get_fields()[field]
+            if field not in all_fields:
+                # Not a nested field
+                continue
+            else:
+                field_serializer = all_fields[field]
+
             if isinstance(field_serializer, Serializer):
-                if isinstance(field_serializer, _ReplaceableField):
+                if isinstance(field_serializer, BaseReplaceableNestedField):
                     value = validated_data.pop(field)
                     fields["foreignkey_related"]["replaceable"] \
                         .update({field: value})
-                elif isinstance(field_serializer, _WritableField):
+                elif isinstance(field_serializer, BaseWritableNestedField):
                     value = validated_data.pop(field)
                     fields["foreignkey_related"]["writable"]\
                         .update({field: value})
             elif (isinstance(field_serializer, ListSerializer) and 
-                    (isinstance(field_serializer, _WritableField) or 
-                    isinstance(field_serializer, _ReplaceableField))):
+                    (isinstance(field_serializer, BaseWritableNestedField) or 
+                    isinstance(field_serializer, BaseReplaceableNestedField))):
 
                 model = self.Meta.model
                 rel = getattr(model, field).rel
@@ -388,24 +601,26 @@ class NestedCreateMixin(object):
         instance = super().create({**validated_data, **foreignkey_related})
         
         self.create_many_to_many_related(
-            instance, 
+            instance,
             fields["many_to"]["many_related"]
         )
 
         self.create_many_to_one_related(
-            instance, 
+            instance,
             fields["many_to"]["one_related"]
         )
         
         return instance
 
 
-class NestedUpdateMixin(object):
+class NestedUpdateMixin(BaseNestedMixin):
     """ Update Mixin """
-    def constrain_error_prefix(self, field):
+    @staticmethod
+    def constrain_error_prefix(field):
         return "Error on %s field: " % (field,)
 
-    def update_replaceable_foreignkey_related(self, instance, data):
+    @staticmethod
+    def update_replaceable_foreignkey_related(instance, data):
         # data format {field: obj}
         objs = {}
         for field, nested_obj in data.items():
@@ -417,27 +632,47 @@ class NestedUpdateMixin(object):
     def update_writable_foreignkey_related(self, instance, data):
         # data format {field: {sub_field: value}}
         objs = {}
+        all_fields = self._get_all_fields()
         for field, values in data.items():
-            # Get serializer class for nested field
-            SerializerClass = type(self.get_fields()[field])
+            # Get nested field serializer
+            serializer = all_fields[field]
+            serializer_class = type(serializer)
+            kwargs = serializer.validation_kwargs
             nested_obj = getattr(instance, field)
-            serializer = SerializerClass(
-                nested_obj, 
-                data=values, 
+            serializer = serializer_class(
+                nested_obj,
+                **kwargs,
+                data=values,
                 context=self.context,
-                partial=self.partial
+                partial=serializer.is_partial
             )
             serializer.is_valid()
-            serializer.save()
-            objs.update({field: nested_obj})
+            if values is None:
+                setattr(instance, field, None)
+                objs.update({field: None})
+            else:
+                obj = serializer.save()
+                if nested_obj is None:
+                    # Patch back newly created object to instance
+                    setattr(instance, field, obj)
+                    objs.update({field: obj})
+                else:
+                    objs.update({field: nested_obj})
         return objs
 
     def bulk_create_many_to_many_related(self, field, nested_obj, data):
-        # Get serializer class for nested field
-        SerializerClass = type(self.get_fields()[field].child)
+        # Get nested field serializer
+        serializer = self._get_all_fields()[field].child
+        serializer_class = type(serializer)
+        kwargs = serializer.validation_kwargs
         pks = []
         for values in data:
-            serializer = SerializerClass(data=values, context=self.context)
+            serializer = serializer_class(
+                **kwargs,
+                data=values,
+                context=self.context,
+                partial=serializer.is_partial
+            )
             serializer.is_valid()
             obj = serializer.save()
             pks.append(obj.pk)
@@ -445,11 +680,18 @@ class NestedUpdateMixin(object):
         return pks
 
     def bulk_create_many_to_one_related(self, field, nested_obj, data):
-        # Get serializer class for nested field
-        SerializerClass = type(self.get_fields()[field].child)
+        # Get nested field serializer
+        serializer = self._get_all_fields()[field].child
+        serializer_class = type(serializer)
+        kwargs = serializer.validation_kwargs
         pks = []
         for values in data:
-            serializer = SerializerClass(data=values, context=self.context)
+            serializer = serializer_class(
+                **kwargs,
+                data=values,
+                context=self.context,
+                partial=serializer.is_partial
+            )
             serializer.is_valid()
             obj = serializer.save()
             pks.append(obj.pk)
@@ -459,15 +701,18 @@ class NestedUpdateMixin(object):
         # {pk: {sub_field: values}}
         objs = []
 
-        # Get serializer class for nested field
-        SerializerClass = type(self.get_fields()[field].child)
+        # Get nested field serializer
+        serializer = self._get_all_fields()[field].child
+        serializer_class = type(serializer)
+        kwargs = serializer.validation_kwargs
         for pk, values in data.items():
             obj = nested_obj.get(pk=pk)
-            serializer = SerializerClass(
-                obj, 
-                data=values, 
-                context=self.context, 
-                partial=self.partial
+            serializer = serializer_class(
+                obj,
+                **kwargs,
+                data=values,
+                context=self.context,
+                partial=serializer.is_partial
             )
             serializer.is_valid()
             obj = serializer.save()
@@ -478,19 +723,22 @@ class NestedUpdateMixin(object):
         # {pk: {sub_field: values}}
         objs = []
 
-        # Get serializer class for nested field
-        SerializerClass = type(self.get_fields()[field].child)
+        # Get nested field serializer
+        serializer = self._get_all_fields()[field].child
+        serializer_class = type(serializer)
+        kwargs = serializer.validation_kwargs
         model = self.Meta.model
         foreignkey = getattr(model, field).field.name
         nested_obj = getattr(instance, field)
         for pk, values in data.items():
             obj = nested_obj.get(pk=pk)
             values.update({foreignkey: instance.pk})
-            serializer = SerializerClass(
-                obj, 
-                data=values, 
-                context=self.context, 
-                partial=self.partial
+            serializer = serializer_class(
+                obj,
+                **kwargs,
+                data=values,
+                context=self.context,
+                partial=serializer.is_partial
             )
             serializer.is_valid()
             obj = serializer.save()
@@ -501,28 +749,28 @@ class NestedUpdateMixin(object):
         # data format {field: {
         # foreignkey_name: name:
         # data: {
-        # ADD: [{sub_field: value}], 
-        # CREATE: [{sub_field: value}], 
+        # ADD: [{sub_field: value}],
+        # CREATE: [{sub_field: value}],
         # REMOVE: [pk],
         # UPDATE: {pk: {sub_field: value}} 
         # }}}
-
         for field, values in data.items():
             nested_obj = getattr(instance, field)
             model = self.Meta.model
             foreignkey = getattr(model, field).field.name
+            all_fields = self._get_all_fields()
             for operation in values:
                 if operation == ADD:
                     pks = values[operation]
-                    model = self.get_fields()[field].child.Meta.model
+                    model = all_fields[field].child.Meta.model
                     qs = model.objects.filter(pk__in=pks)
                     qs.update(**{foreignkey: instance.pk})
                 elif operation == CREATE:
                     for v in values[operation]:
                         v.update({foreignkey: instance.pk})
                     self.bulk_create_many_to_one_related(
-                        field, 
-                        nested_obj, 
+                        field,
+                        nested_obj,
                         values[operation]
                     )
                 elif operation == REMOVE:
@@ -530,7 +778,7 @@ class NestedUpdateMixin(object):
                     qs.filter(pk__in=values[operation]).delete()
                 elif operation == UPDATE:
                     self.bulk_update_many_to_one_related(
-                        field, 
+                        field,
                         instance,
                         values[operation]
                     )
@@ -543,8 +791,8 @@ class NestedUpdateMixin(object):
 
     def update_many_to_many_related(self, instance, data):
         # data format {field: {
-        # ADD: [{sub_field: value}], 
-        # CREATE: [{sub_field: value}], 
+        # ADD: [{sub_field: value}],
+        # CREATE: [{sub_field: value}],
         # REMOVE: [pk],
         # UPDATE: {pk: {sub_field: value}} 
         # }}
@@ -557,11 +805,11 @@ class NestedUpdateMixin(object):
                         nested_obj.add(*pks)
                     except Exception as e:
                         msg = self.constrain_error_prefix(field) + str(e)
-                        raise ValidationError(msg)
+                        raise ValidationError(msg) from None
                 elif operation == CREATE:
                     self.bulk_create_many_to_many_related(
-                        field, 
-                        nested_obj, 
+                        field,
+                        nested_obj,
                         values[operation]
                     )
                 elif operation == REMOVE:
@@ -570,11 +818,11 @@ class NestedUpdateMixin(object):
                         nested_obj.remove(*pks)
                     except Exception as e:
                         msg = self.constrain_error_prefix(field) + str(e)
-                        raise ValidationError(msg)
+                        raise ValidationError(msg) from None
                 elif operation == UPDATE:
                     self.bulk_update_many_to_many_related(
-                        field, 
-                        nested_obj, 
+                        field,
+                        nested_obj,
                         values[operation]
                     )
                 else:
@@ -599,20 +847,26 @@ class NestedUpdateMixin(object):
         # Make a partal copy of validated_data so that we can
         # iterate and alter it
         data = copy.copy(validated_data)
+        all_fields = self._get_all_fields()
         for field in data:
-            field_serializer = self.get_fields()[field]
+            # Not a nested field
+            if field not in all_fields:
+                continue
+            else:
+                field_serializer = all_fields[field]
+            
             if isinstance(field_serializer, Serializer):
-                if isinstance(field_serializer, _ReplaceableField):
+                if isinstance(field_serializer, BaseReplaceableNestedField):
                     value = validated_data.pop(field)
                     fields["foreignkey_related"]["replaceable"] \
                         .update({field: value})
-                elif isinstance(field_serializer, _WritableField):
+                elif isinstance(field_serializer, BaseWritableNestedField):
                     value = validated_data.pop(field)
                     fields["foreignkey_related"]["writable"] \
                         .update({field: value})
             elif (isinstance(field_serializer, ListSerializer) and
-                    (isinstance(field_serializer, _WritableField) or 
-                    isinstance(field_serializer, _ReplaceableField))):
+                    (isinstance(field_serializer, BaseWritableNestedField) or 
+                    isinstance(field_serializer, BaseReplaceableNestedField))):
                 model = self.Meta.model
                 rel = getattr(model, field).rel
     
@@ -645,4 +899,4 @@ class NestedUpdateMixin(object):
             fields["many_to"]["one_related"]
         )
 
-        return super().update(instance, validated_data)
\ No newline at end of file
+        return super().update(instance, validated_data)
